// Copyright (C) 2018-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package xbtree

// ΔBtail provides BTree-level history tail.
//
// It translates ZODB object-level changes to information about which keys of
// which BTree were modified, and provides service to query that information.
// See ΔBtail class documentation for usage details.
//
//
// ΔBtail organization
//
// ΔBtail keeps raw ZODB history in ΔZtail and uses BTree-diff algorithm(*) to
// turn δZ into BTree-level diff. For each tracked BTree a separate ΔTtail is
// maintained with tree-level history in ΔTtail.vδT .
//
// Because it is very computationally expensive(+) to find out for an object to
// which BTree it belongs, ΔBtail cannot provide full BTree-level history given
// just ΔZtail with δZ changes. Because of this ΔBtail requires help from
// users, which are expected to call ΔBtail.Track(treepath) to let ΔBtail know
// that such and such ZODB objects constitute a path from root of a tree to some
// of its leaf. After Track call the objects from the path and tree keys, that
// are covered by leaf node, become tracked: from now-on ΔBtail will detect
// and provide BTree-level changes caused by any change of tracked tree objects
// or tracked keys. This guarantee can be provided because ΔBtail now knows
// that such and such objects belong to a particular tree.
//
// To manage knowledge which tree part is tracked ΔBtail uses PPTreeSubSet.
// This data-structure represents so-called PP-connected set of tree nodes:
// simply speaking it builds on some leafs and then includes parent(leaf),
// parent(parent(leaf)), etc. In other words it's a "parent"-closure of the
// leafs. The property of being PP-connected means that starting from any node
// from such set, it is always possible to reach root node by traversing
// .parent links, and that every intermediate node went-through during
// traversal also belongs to the set.
//
// A new Track request potentially grows tracked keys coverage. Due to this,
// ΔBtail needs to recompute potentially whole vδT of the affected tree. This
// recomputation is managed by "rebuild..." family of functions and uses the
// same treediff algorithm, that Update is using, but modulo PPTreeSubSet
// corresponding to δ key coverage. Update also potentially needs to rebuild
// whole vδT history, not only append new δT, because a change to tracked tree
// nodes can result in growth of tracked key coverage.
//
// Queries are relatively straightforward code that work on vδT snapshot. The
// main complexity, besides BTree-diff algorithm, lies in recomputing vδT when
// set of tracked keys changes.  XXX and in concurrency
//
//
// XXX concurrency
//
//
// --------
//
// (*) implemented in treediff.go
// (+) full database scan

import (
	"context"
	"fmt"
	"sort"
	"strings"

	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"

	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xbtree/blib"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xtail"
)

const traceΔBtail = false
const debugΔBtail = false


// ΔBtail represents tail of revisional changes to BTrees.
//
// It semantically consists of
//
//	[]δB			; rev ∈ (tail, head]
//
// where δB represents a change in BTrees space
//
//	δB:
//		.rev↑
//		{} root -> {}(key, δvalue)
//
// It covers only changes to keys from tracked subset of BTrees parts.
// In particular a key that was not explicitly requested to be tracked, even if
// it was changed in δZ, is not guaranteed to be present in δB.
//
// ΔBtail provides the following operations:
//
//   .Track(path)	- start tracking tree nodes and keys; root=path[0], keys=path[-1].(lo,hi]
//
//   .Update(δZ) -> δB				- update BTree δ tail given raw ZODB changes
//   .ForgetPast(revCut)			- forget changes ≤ revCut
//   .SliceByRev(lo, hi) -> []δB		- query for all trees changes with rev ∈ (lo, hi]
//   .SliceByRootRev(root, lo, hi) -> []δT	- query for changes of a tree with rev ∈ (lo, hi]
//   .GetAt(root, key, at) -> (value, rev)	- get root[key] @at assuming root[key] ∈ tracked
//
// where δT represents a change to one tree
//
//	δT:
//		.rev↑
//		{}(key, δvalue)
//
// An example for tracked set is a set of visited BTree paths.
// There is no requirement that tracked set belongs to only one single BTree.
//
// XXX concurrent use
//
// See also zodb.ΔTail and zdata.ΔFtail
type ΔBtail struct {
	// raw ZODB changes; Kept to rebuild .byRoot after new Track.
	// includes all changed objects, not only tracked ones.
	δZtail *zodb.ΔTail

	vδBroots  []_ΔBroots         // [] (rev↑, roots changed in this rev)
	byRoot map[zodb.Oid]*_ΔTtail // {} root -> [] k/v change history; only for keys ∈ tracked subset

	// set of tracked nodes as of @head state.
	// For this set all vδT are fully computed.
	// The set of nodes that were requested to be tracked, but were not yet
	// taken into account, is kept in _ΔTtail.trackNew & co.
	trackSet blib.PPTreeSubSet

	// set of trees for which _ΔTtail.trackNew is non-empty
	trackNewRoots setOid

	// handle to make connections to access database.
	// TODO allow client to optionally provide zconnOld/zconnNew on e.g. Update()
	db *zodb.DB // to open connections to load new/old tree|buckets
}

// _ΔTtail represent tail of revisional changes to one BTree.
//
// See ΔBtail documentation for details.
type _ΔTtail struct {
	vδT []ΔTree // changes to tree keys; rev↑. covers keys ∈ tracked subset

	// set of nodes that were requested to be tracked in this tree, but for
	// which vδT was not yet rebuilt
	trackNew blib.PPTreeSubSet

	// XXX + trackNewKeys RangedKeySet (concurrency)
	// XXX + trackSetKeys RangedKeySet
}

// _ΔBroots represents roots-only part of ΔB.
//
// It describes which trees were changed, but does not provide δkv details for changed trees.
type _ΔBroots struct {
	Rev   zodb.Tid
	Roots setOid   // which roots changed in this revision
}

// ΔB represents a change in BTrees space.
type ΔB struct {
	Rev    zodb.Tid
	ByRoot map[zodb.Oid]map[Key]ΔValue // {} root -> {}(key, δvalue)
}

// ΔTree describes changes to one BTree in one revision.
type ΔTree struct {
	Rev zodb.Tid
	KV  map[Key]ΔValue
}


// NewΔBtail creates new empty ΔBtail object.
//
// Initial tracked set is empty.
// Initial coverage is (at₀, at₀].
//
// db will be used by ΔBtail to open database connections to load data from
// ZODB when needed.
func NewΔBtail(at0 zodb.Tid, db *zodb.DB) *ΔBtail {
	return &ΔBtail{
		δZtail:        zodb.NewΔTail(at0),
		vδBroots:      nil,
		byRoot:        map[zodb.Oid]*_ΔTtail{},
		trackSet:      blib.PPTreeSubSet{},
		trackNewRoots: setOid{},
		db:            db,
	}
}

// newΔTtail creates new empty _ΔTtail object.
func newΔTtail() *_ΔTtail {
	return &_ΔTtail{
		trackNew:  blib.PPTreeSubSet{},
	}
}

// Clone returns copy of ΔBtail.
func (orig *ΔBtail) Clone() *ΔBtail {
	klon := &ΔBtail{db: orig.db}

	// δZtail
	klon.δZtail = zodb.NewΔTail(orig.Tail())
	for _, δZ := range orig.δZtail.Data() {
		klon.δZtail.Append(δZ.Rev, δZ.Changev)
	}

	// vδBroots
	klon.vδBroots = make([]_ΔBroots, 0, len(orig.vδBroots))
	for _, origδBroots := range orig.vδBroots {
		klonδBroots := _ΔBroots{
			Rev:   origδBroots.Rev,
			Roots: origδBroots.Roots.Clone(),
		}
		klon.vδBroots = append(klon.vδBroots, klonδBroots)
	}

	// byRoot
	klon.byRoot = make(map[zodb.Oid]*_ΔTtail, len(orig.byRoot))
	for root, origΔTtail := range orig.byRoot {
		klon.byRoot[root] = origΔTtail.Clone()
	}

	// trackSet, trackNewRoots
	klon.trackSet = orig.trackSet.Clone()
	klon.trackNewRoots = orig.trackNewRoots.Clone()

	return klon
}

// Clone returns copy of _ΔTtail.
func (orig *_ΔTtail) Clone() *_ΔTtail {
	klon := &_ΔTtail{}
	klon.vδT = vδTClone(orig.vδT)
	klon.trackNew = orig.trackNew.Clone()
	return klon
}

// vδTClone returns deep copy of []ΔTree.
func vδTClone(orig []ΔTree) []ΔTree {
	klon := make([]ΔTree, 0, len(orig))
	for _, origδT := range orig {
		klonδT := ΔTree{
			Rev: origδT.Rev,
			KV:  make(map[Key]ΔValue, len(origδT.KV)),
		}
		for k, δv := range origδT.KV {
			klonδT.KV[k] = δv
		}
		klon = append(klon, klonδT)
	}
	return klon
}

// (tail, head] coverage
func (δBtail *ΔBtail) Head() zodb.Tid { return δBtail.δZtail.Head() }
func (δBtail *ΔBtail) Tail() zodb.Tid { return δBtail.δZtail.Tail() }


// ---- Track/rebuild/Update/Forget ----

// Track adds tree path to tracked set.
//
// path[0] signifies tree root.
// All path elements must be Tree except last one which, for non-empty tree, must be Bucket.
//
// Besides key (which might point to value or hole), δBtail will also track all
// keys covered by leaf node. In particular after request for KeyMax or KeyMin
// to be tracked, δBtail will keep on tracking changes to maximum or minimum
// keys correspondingly.
//
// Objects in path must be with .PJar().At() == .Head()
func (δBtail *ΔBtail) Track(key Key, nodePath []Node) {
	// NOTE key not needed for anything besides tracing
	// (tracking set will be added with all keys, covered by leaf keyrange)

	if traceΔBtail {
		pathv := []string{}
		for _, node := range nodePath { pathv = append(pathv, vnode(node)) }
		tracefΔBtail("\nTrack [%v] %s\n", key, strings.Join(pathv, " -> "))
		tracefΔBtail("trackSet: %s\n", δBtail.trackSet) // XXX locking
	}

	head := δBtail.Head()
	for _, node := range nodePath {
		nodeAt := node.PJar().At()
		if nodeAt != head {
			panicf("node.at (@%s)  !=  δBtail.head (@%s)", nodeAt, head)
		}
	}

	path := nodePathToPath(nodePath)
	δBtail.track(key, path)
}

// nodePathToPath converts path from []Node to []Oid.
func nodePathToPath(nodePath []Node) (path []zodb.Oid) {
	// assert nodePath = Tree Tree ... Tree Bucket
	l := len(nodePath)
	switch {
	case l == 0:
		panic("empty path")
	case l == 1:
		// must be empty Tree
		_ = nodePath[0].(*Tree)
	default:
		// must be Tree Tree ... Tree Bucket
		for _, node := range nodePath[:l-1] {
			_ = node.(*Tree)
		}
		_ = nodePath[l-1].(*Bucket)
	}

	path = make([]zodb.Oid, l)
	for i, node := range nodePath {
		path[i] = node.POid()
	}
	return path
}

func (δBtail *ΔBtail) track(key Key, path []zodb.Oid) {
	// XXX locking

	// first normalize path: remove embedded bucket and check if it was an
	// empty artificial tree. We need to do the normalization because we
	// later check whether leaf path[-1] ∈ trackSet and without
	// normalization path[-1] can be InvalidOid.
	path = blib.NormPath(path)
	if len(path) == 0 {
		return // empty tree
	}

	root := path[0]

	// nothing to do if key is already tracked
	leaf := path[len(path)-1]
	if δBtail.trackSet.Has(leaf) {
		tracefΔBtail("->T: nop\n")
		path_ := δBtail.trackSet.Path(leaf)
		if !pathEqual(path, path_) {
			panicf("BUG: key %s is already tracked via path=%v\ntrack requests path=%v", kstr(key), path_, path)
		}
		return
	}

	// queue path into trackNew
	δTtail, ok := δBtail.byRoot[root]
	if !ok {
		δTtail = newΔTtail()
		δBtail.byRoot[root] = δTtail
	}

	δBtail.trackNewRoots.Add(root)
	δTtail.trackNew.AddPath(path)
	tracefΔBtail("->T: [%s].trackNew -> %s\n", root, δTtail.trackNew)
}

// rebuildAll rebuilds ΔBtail taking all trackNew requests into account.
func (δBtail *ΔBtail) rebuildAll() (err error) {
	defer xerr.Context(&err, "ΔBtail rebuildAll")
	// XXX locking

	tracefΔBtail("\nRebuildAll @%s..@%s    trackNewRoots: %s\n", δBtail.Tail(), δBtail.Head(), δBtail.trackNewRoots)

	for root := range δBtail.trackNewRoots {
		delete(δBtail.trackNewRoots, root)
		δBtail.rebuild1(root)
	}

	return nil
}

// rebuild1IfNeeded rebuilds ΔBtail for single root if that root needs rebuilding.
func (δBtail *ΔBtail) rebuild1IfNeeded(root zodb.Oid) error {
	// XXX locking
	_, ok := δBtail.trackNewRoots[root]
	if !ok {
		return nil
	}

	delete(δBtail.trackNewRoots, root)
	return δBtail.rebuild1(root)
}

// rebuild1 rebuilds ΔBtail for single root.
func (δBtail *ΔBtail) rebuild1(root zodb.Oid) error {
	// XXX locking
	δTtail := δBtail.byRoot[root] // must be there
	δtrackSet, δrevSet, err := δTtail.rebuild(root, δBtail.δZtail, δBtail.db)
	if err != nil {
		return err
	}
	δBtail.trackSet.UnionInplace(δtrackSet)
	δBtail.vδBroots_Update(root, δrevSet)
	return nil
}


// rebuild rebuilds _ΔTtail taking trackNew requests into account.
//
// It returns:
//
// - set of nodes that must be added to ΔBtail.trackSet to account for
//   keys that becomes tracked. Note: this set is potentially wider compared to what was in .trackNew.
// - set of revisions for which new entries in .vδT have been created.
func (δTtail *_ΔTtail) rebuild(root zodb.Oid, δZtail *zodb.ΔTail, db *zodb.DB) (δtrackSet blib.PPTreeSubSet, δrevSet setTid, err error) {
	defer xerr.Contextf(&err, "ΔTtail<%s> rebuild", root)
	// XXX locking

	tracefΔBtail("\nRebuild %s @%s .. @%s\n", root, δZtail.Tail(), δZtail.Head())
	tracefΔBtail("trackNew:       %v\n", δTtail.trackNew)

	trackNew := δTtail.trackNew
	δTtail.trackNew = blib.PPTreeSubSet{}

	if len(trackNew) == 0 {
		return nil, nil, nil
	}

	δrevSet = setTid{}

	// clone vδT before modifying it
	// queries such as SliceByRootRev return slices of vδT and we do not
	// want to change data that is already returned to user.
	δTtail.vδT = vδTClone(δTtail.vδT)

	// go backwards and merge vδT <- treediff(lo..hi/trackNew)
	vδZ := δZtail.Data()
	for {
		δtkeycov := &blib.RangedKeySet{} // all keys coming into tracking set during this lo<-hi scan
		trackNewCur := trackNew.Clone()  // trackNew adjusted as of when going to i<- entry
		for i := len(vδZ)-1; i>=0; i-- {
			δZ := vδZ[i]

			var atPrev zodb.Tid
			if i > 0 {
				atPrev = vδZ[i-1].Rev
			} else {
				atPrev = δZtail.Tail()
			}

			δtrackNew, δtkeycov_, newRevEntry, err := δTtail.rebuild1(atPrev, δZ, trackNewCur, db)
			if err != nil {
				return nil, nil, err
			}

			trackNewCur.ApplyΔ(δtrackNew)
			δtkeycov.UnionInplace(δtkeycov_)
			if newRevEntry {
				δrevSet.Add(δZ.Rev)
			}
		}

		// an iteration closer to tail may turn out to add a key to the tracking set.
		// We have to recheck all entries newer that revision for changes to that key,
		// for example:
		//
		//              8        5*
		//             / \  <-  / \
		//            2   8    2*  7
		//
		//     here initial tracked set is 5*-2*. Going to earlier revision
		//     2'th keycov range is widen from [-∞,5) to [-∞,7), so 5*-7 in
		//     later revision have to be rechecked because 7 was added into
		//     tracking set.
		//
		// Implement this via restarting from head and cycling until
		// set of tracked keys does not grow anymore.
		if δtkeycov.Empty() {
			break
		}

		err := widenTrackNew(trackNew, δtkeycov, root, δZtail.Head(), db)
		if err != nil {
			return nil, nil, err
		}
	}

	return trackNew, δrevSet, nil
}

// rebuild1 rebuilds δT for single δZ.
//
// δtrackNew/δtkeycov represents how trackNew changes when going through `atPrev <- δZ.Rev` .
// newRevEntry indicates whether δZ.Rev was not there before in .vδT and new corresponding δT entry was created.
func (δTtail *_ΔTtail) rebuild1(atPrev zodb.Tid, δZ zodb.ΔRevEntry, trackNew blib.PPTreeSubSet, db *zodb.DB) (δtrackNew *blib.ΔPPTreeSubSet, δtkeycov *blib.RangedKeySet, newRevEntry bool, err error) {
	defer xerr.Contextf(&err, "rebuild1 %s<-%s", atPrev, δZ.Rev)

	debugfΔBtail("\n  rebuild1 @%s <- @%s\n", atPrev, δZ.Rev)
	debugfΔBtail("  δZ:\t%v\n", δZ.Changev)
	debugfΔBtail("  trackNew:  %v\n", trackNew)
	defer func() {
		debugfΔBtail("-> δtrackNew:   %v\n", δtrackNew)
		debugfΔBtail("-> δtkeycov:    %v\n", δtkeycov)
		debugfΔBtail("-> newRevEntry: %v\n", newRevEntry)
		debugfΔBtail("\n\n")
	}()

	// NOTE: keep vvv in sync with ΔBtail._Update1

	δZTC, δtopsByRoot := δZConnectTracked(δZ.Changev, trackNew)

	// skip opening DB connections if there is no change to this tree
	if len(δtopsByRoot) == 0 {
		return blib.NewΔPPTreeSubSet(), &blib.RangedKeySet{}, false, nil
	}

	if len(δtopsByRoot) != 1 {
		panicf("BUG: δtopsByRoot has > 1 entries: %v\ntrackNew: %v\nδZ: %v", δtopsByRoot, trackNew, δZ)
	}
	var root  zodb.Oid
	var δtops setOid
	for root_, δtops_ := range δtopsByRoot {
		root  = root_
		δtops = δtops_
	}


	// open ZODB connection corresponding to "current" and "prev" states
	txn, ctx := transaction.New(context.TODO()) // TODO - merge in cancel via ctx arg
	defer txn.Abort()

	zconnPrev, err := db.Open(ctx, &zodb.ConnOptions{At: atPrev})
	if err != nil {
		return nil, nil, false, err
	}
	zconnCurr, err := db.Open(ctx, &zodb.ConnOptions{At: δZ.Rev})
	if err != nil {
		return nil, nil, false, err
	}

	// diff backwards curr -> prev
	δT, δtrack, δtkeycov, err := treediff(ctx, root, δtops, δZTC, trackNew, zconnCurr, zconnPrev)
	if err != nil {
		return nil, nil, false, err
	}

	debugfΔBtail("  -> root<%s>  δkv*: %v  δtrack*: %v  δtkeycov*: %v\n", root, δT, δtrack, δtkeycov)

	if len(δT) == 0 { // an object might be resaved without change
		return δtrack, δtkeycov, false, nil
	}

	// δTtail.vδT <- merge δT*
	l := len(δTtail.vδT)
	j := sort.Search(l, func(k int) bool {
		return δZ.Rev <= δTtail.vδT[k].Rev
	})
	if j == l || δTtail.vδT[j].Rev != δZ.Rev {
		newRevEntry = true
		δTcurr := ΔTree{Rev: δZ.Rev, KV: map[Key]ΔValue{}}
		// insert(@j, δTcurr)
		δTtail.vδT = append(δTtail.vδT[:j],
				    append([]ΔTree{δTcurr},
				    δTtail.vδT[j:]...)...)
	}
	δTcurr := δTtail.vδT[j]

	for k, δv := range δT {
		// the diff was backward; δTtail entries are with diff forward
		δv.New, δv.Old = δv.Old, δv.New

		δv_, already := δTcurr.KV[k]
		if already {
			if δv != δv_ {
				panicf("[%v] inconsistent δv:\nδTcurr: %v\nδT:  %v", k, δTcurr, δT)
			}
		} else {
			δTcurr.KV[k] = δv
		}
	}

	return δtrack, δtkeycov, newRevEntry, nil
}

// widenTrackNew widens trackNew to cover δtkeycov.
func widenTrackNew(trackNew blib.PPTreeSubSet, δtkeycov *blib.RangedKeySet, root zodb.Oid, at zodb.Tid, db *zodb.DB) (err error) {
	defer xerr.Contextf(&err, "widenTrackNew tree<%s> @%s +%s", root, at, δtkeycov)
	debugfΔBtail("\n  widenTrackNew %s @%s +%s", root, at, δtkeycov)

	txn, ctx := transaction.New(context.TODO()) // TODO - merge in cancel via ctx arg
	defer txn.Abort()

	zhead, err := db.Open(ctx, &zodb.ConnOptions{At: at});  /*X*/ if err != nil { return err }
	xtree, err := zgetNodeOrNil(ctx, zhead, root);          /*X*/ if err != nil { return err }
	if xtree == nil {
		// root deleted -> root node covers [-∞,∞)
		trackNew.AddPath([]zodb.Oid{root})
		return nil
	}
	tree := xtree.(*Tree) // must succeed

	top := &nodeInRange{prefix: nil, keycov: blib.KeyRange{KeyMin, KeyMax}, node: tree}
	V   := rangeSplit{top}
	for _, r := range δtkeycov.AllRanges() {
		lo := r.Lo
		for {
			b, err := V.GetToLeaf(ctx, lo);  /*X*/ if err != nil { return err }
			trackNew.AddPath(b.Path())

			// continue with next right bucket until r coverage is complete
			if r.Hi_ <= b.keycov.Hi_ {
				break
			}
			lo = b.keycov.Hi_ + 1
		}
	}
	return nil
}


// Update updates δB with object-level ZODB changes.
//
// Only those objects from δZ that belong to tracked set are guaranteed to be
// taken into account. In other words a tree history will assuredly include
// only those keys, that correspond to tracked subset of δZ.
//
// δZ must include all objects changed by ZODB transaction.
//
// TODO optionally accept zconnOld/zconnNew from client
func (δBtail *ΔBtail) Update(δZ *zodb.EventCommit) (_ ΔB, err error) {
	headOld := δBtail.Head()
	defer xerr.Contextf(&err, "ΔBtail.Update %s -> %s", headOld, δZ.Tid)

	δB1, err := δBtail._Update1(δZ)

	δB := ΔB{Rev: δZ.Tid, ByRoot: make(map[zodb.Oid]map[Key]ΔValue)}
	for root, δT1 := range δB1.ByRoot {
		δTtail := δBtail.byRoot[root] // must succeed

		// δtkeycov1 != ø   ->  rebuild δTtail with trackNew ~= δtkeycov1
		if !δT1.δtkeycov1.Empty() && δBtail.δZtail.Len() > 1 {
			trackNew := blib.PPTreeSubSet{}
			err := widenTrackNew(trackNew, δT1.δtkeycov1, root, δBtail.Head(), δBtail.db)
			if err != nil {
				return ΔB{}, err
			}
			if !δTtail.trackNew.Empty() {
				panicf("BUG: ΔBtail.Update: ΔTtail<%s>.trackNew != ø after _Update1", root)
			}
			δTtail.trackNew = trackNew

			// NOTE we cannot skip computing diff for HEAD~..HEAD
			// even after _Update1 because _Update1 was working with different trackNew.
			δtrackSet, δrevSet, err := δTtail.rebuild(root, δBtail.δZtail, δBtail.db)
			if err != nil {
				return ΔB{}, err
			}
			δBtail.trackSet.UnionInplace(δtrackSet)
			δBtail.vδBroots_Update(root, δrevSet)
		}

		// build δB. Even if δT=ø after _Update1, but δtkeycov1 != ø, above
		// rebuild could result in head δT becoming != ø. Detect that δTtail head
		// is anew by comparing to δZ.Rev.
		l := len(δTtail.vδT)
		if l > 0 {
			δT := δTtail.vδT[l-1] // δT head
			if δT.Rev == δZ.Tid {
				δB.ByRoot[root] = δT.KV
			}
		}
	}

	// vδBroots += δB  (δB.Rev could be already there added by ^^^ rebuild)
	for root := range δB.ByRoot {
		δBtail.vδBroots_Update1(root, δB.Rev)
	}

	return δB, err
}

// _Update1 serves Update and performs direct update of δTtail head elements from δZ.
// On key coverage growth rebuilding tail of the history is done by Update itself.
//
// _Update1 is also used in tests to verify δtkeycov return from treediff.
type _ΔBUpdate1 struct {
	ByRoot map[zodb.Oid]*_ΔTUpdate1
}
type _ΔTUpdate1 struct {
	δtkeycov1 *blib.RangedKeySet // {} root -> δtrackedKeys after first treediff (always grow)
}
func (δBtail *ΔBtail) _Update1(δZ *zodb.EventCommit) (δB1 _ΔBUpdate1, err error) {
	headOld := δBtail.Head()
	defer xerr.Contextf(&err, "ΔBtail.update1 %s -> %s", headOld, δZ.Tid)

	tracefΔBtail("\nUpdate  @%s -> @%s    δZ: %v\n", δBtail.Head(), δZ.Tid, δZ.Changev)
	tracefΔBtail("trackSet:       %v\n", δBtail.trackSet)
	for _, root := range δBtail.trackNewRoots.SortedElements() {
		δTtail := δBtail.byRoot[root]
		tracefΔBtail("[%s].trackNew:  %v\n", root, δTtail.trackNew)
	}

	δB1 = _ΔBUpdate1{ByRoot: make(map[zodb.Oid]*_ΔTUpdate1)}

	// update .trackSet and vδB from .trackNew
	err = δBtail.rebuildAll()
	if err != nil {
		return δB1, err
	}

	δBtail.δZtail.Append(δZ.Tid, δZ.Changev)

	// NOTE: keep vvv in sync with ΔTtail.rebuild1

	δZTC, δtopsByRoot := δZConnectTracked(δZ.Changev, δBtail.trackSet)

	// skip opening DB connections if there is no change to any tree node
	if len(δtopsByRoot) == 0 {
		return δB1, nil
	}

	// open ZODB connections corresponding to "old" and "new" states
	// TODO caller should provide one of those (probably new) as usually it has it
	txn, ctx := transaction.New(context.TODO()) // TODO - merge in cancel via ctx arg
	defer txn.Abort()
	zconnOld, err := δBtail.db.Open(ctx, &zodb.ConnOptions{At: headOld})
	if err != nil {
		return δB1, err
	}
	zconnNew, err := δBtail.db.Open(ctx, &zodb.ConnOptions{At: δZ.Tid})
	if err != nil {
		return δB1, err
	}

	for root, δtops := range δtopsByRoot {
		δT, δtrack, δtkeycov, err := treediff(ctx, root, δtops, δZTC, δBtail.trackSet, zconnOld, zconnNew)
		if err != nil {
			return δB1, err
		}

		tracefΔBtail("\n-> root<%s>  δkv: %v  δtrack: %v  δtkeycov: %v\n", root, δT, δtrack, δtkeycov)

		δTtail := δBtail.byRoot[root] // must be there
		if len(δT) > 0 { // an object might be resaved without change
			// NOTE no need to clone .vδT here because we only append to it:
			//      Even though queries return vδT aliases, append
			//      cannot change any slice returned by query to users.
			δTtail.vδT = append(δTtail.vδT, ΔTree{Rev: δZ.Tid, KV: δT})
		}

		δBtail.trackSet.ApplyΔ(δtrack)
		δB1.ByRoot[root] = &_ΔTUpdate1{δtkeycov1: δtkeycov}
	}

	return δB1, nil
}

// vδBroots_Update updates .vδBroots to remember that _ΔTtail for root has
// changed entries with δrevSet revisions.
func (δBtail *ΔBtail) vδBroots_Update(root zodb.Oid, δrevSet setTid) {
	// XXX locking
	// TODO δrevSet -> []rev↑  and  merge them in one go
	for rev := range δrevSet {
		δBtail.vδBroots_Update1(root, rev)
	}
}

func (δBtail *ΔBtail) vδBroots_Update1(root zodb.Oid, rev zodb.Tid) {
	l := len(δBtail.vδBroots)
	j := sort.Search(l, func(k int) bool {
		return rev <= δBtail.vδBroots[k].Rev
	})
	if j == l || rev != δBtail.vδBroots[j].Rev {
		δBroots := _ΔBroots{Rev: rev, Roots: setOid{}}
		// insert(@j, δBroots)
		δBtail.vδBroots = append(δBtail.vδBroots[:j],
					 append([]_ΔBroots{δBroots},
					 δBtail.vδBroots[j:]...)...)
	}
	δBroots := δBtail.vδBroots[j]
	δBroots.Roots.Add(root)
}

// ForgetPast forgets history entries with revision ≤ revCut.
func (δBtail *ΔBtail) ForgetPast(revCut zodb.Tid) {
	// XXX locking

	δBtail.δZtail.ForgetPast(revCut)

	// go through vδBroots till revcut -> find which trees to trim -> trim ΔTtails.

	totrim := setOid{} // roots whose _ΔTtail has changes ≤ revCut
	icut := 0
	for ; icut < len(δBtail.vδBroots); icut++ {
		δBroots := δBtail.vδBroots[icut]
		if δBroots.Rev > revCut {
			break
		}
		totrim.Update(δBroots.Roots)
	}

	// vδBroots[:icut] should be forgotten
	δBtail.vδBroots = append([]_ΔBroots(nil), δBtail.vδBroots[icut:]...)

	// trim roots
	for root := range totrim {
		δTtail := δBtail.byRoot[root] // must be present
		δTtail.forgetPast(revCut)
	}
}

func (δTtail *_ΔTtail) forgetPast(revCut zodb.Tid) {
	// XXX locking

	icut := 0
	for ; icut < len(δTtail.vδT); icut++ {
		if δTtail.vδT[icut].Rev > revCut {
			break
		}
	}

	// vδT[:icut] should be forgotten
	// NOTE clones vδT because queries return vδT aliases
	δTtail.vδT = append([]ΔTree(nil), δTtail.vδT[icut:]...)
}


// ---- queries ----

// GetAt tries to retrieve root[key]@at from δBtail data.
//
// If δBtail has δB entry that covers root[key]@at, corresponding value
// (VDEL means deletion) and valueExact=true are returned. If δBtail data
// allows to determine revision of root[key]@at value, corresponding revision
// and revExact=true are returned. If revision of root[key]@at cannot be
// determined (rev=δBtail.Tail, revExact=false) are returned.
//
// In particular:
//
// If δBtail has no δB entry that covers root[key]@at, return is
//
//	value:      VDEL,
//	valueExact: false,
//	rev:        δBtail.Tail,
//	revExact:   false
//
// If δBtail has    δB entry that covers root[key]@at, return is
//
//	value:      δB.δvalue.New,
//	valueExact: true,
//	rev:        δB.rev,
//	revExact:   true
//
// If δBtail has    δB entry that covers value for root[key]@at via
// δB.δvalue.Old, but not entry that covers root[key]@at fully, return is:
//
//	value:      δB.δvalue.Old,
//	valueExact: true,
//	rev:        δBtail.Tail,
//	revExact:   false
//
// key must be tracked
// at  must ∈ (tail, head]
func (δBtail *ΔBtail) GetAt(root zodb.Oid, key Key, at zodb.Tid) (value Value, rev zodb.Tid, valueExact, revExact bool, err error) {
	defer xerr.Contextf(&err, "δBtail: root<%s>: get %d @%s", root, key, at)

	tail := δBtail.Tail()
	head := δBtail.Head()
	if !(tail < at && at <= head) {
		panicf("at out of bounds: at: @%s,  (tail, head] = (@%s, @%s]", at, tail, head)
	}

	// XXX locking

	value = VDEL
	valueExact = false
	rev = tail
	revExact = false

	// XXX need to rebuild only if key was not rebuilt yet
	// XXX need to rebuild only for key, not for whole trackNew
	err = δBtail.rebuild1IfNeeded(root)
	if err != nil {
		return value, rev, valueExact, revExact, err
	}

	δTtail := δBtail.byRoot[root]
	if δTtail == nil {
		panicf("δBtail: root<%s> not tracked", root)
	}
	// TODO key not tracked	-> panic  (check key ∈ lastRevOf  -- see vvv)

	// TODO -> index lastRevOf(key) | linear scan ↓ looking for change ≤ at
	for i := len(δTtail.vδT)-1; i >= 0; i-- {
		δT := δTtail.vδT[i]
		δvalue, ok_ := δT.KV[key]
		if ok_ {
			valueExact = true
			if δT.Rev > at {
				value = δvalue.Old
			} else {
				value = δvalue.New
				rev = δT.Rev
				revExact = true
				break
			}
		}
	}

	return value, rev, valueExact, revExact, nil
}

// TODO if needed
// func (δBtail *ΔBtail) SliceByRev(lo, hi zodb.Tid) /*readonly*/ []ΔB

// SliceByRootRev returns history of a tree changes in (lo, hi] range.
//
// it must be called with the following condition:
//
//	tail ≤ lo ≤ hi ≤ head
//
// the caller must not modify returned slice.
//
// Only tracked keys are guaranteed to be present.
//
// Note: contrary to regular go slicing, low is exclusive while high is inclusive.
func (δBtail *ΔBtail) SliceByRootRev(root zodb.Oid, lo, hi zodb.Tid) /*readonly*/[]ΔTree {
	xtail.AssertSlice(δBtail, lo, hi)
	// XXX locking

	err := δBtail.rebuild1IfNeeded(root)
	if err != nil {
		panic(err) // XXX
	}

	δTtail, ok := δBtail.byRoot[root]
	if !ok {
		return []ΔTree{}
	}

	vδT := δTtail.vδT
	l := len(vδT)
	if l == 0 {
		return nil
	}

	// find max j : [j].rev ≤ hi		linear scan -> TODO binary search
	j := l - 1
	for ; j >= 0 && vδT[j].Rev > hi; j-- {}
	if j < 0 {
		return nil // ø
	}

	// find max i : [i].rev > lo		linear scan -> TODO binary search
	i := j
	for ; i >= 0 && vδT[i].Rev > lo; i-- {}
        i++

	// NOTE: no need to duplicate returned vδT slice because
	// _ΔTtail.rebuild clones vδT before modifying it. This way the data we
	// return to caller will stay unchanged even if rebuild is running
	// simultaneously.
	return vδT[i:j+1]
}


// ----------------------------------------


// ΔZtail returns raw ZODB changes accumulated in δBtail so far.
//
// the caller must not modify returned δZtail.
func (δBtail *ΔBtail) ΔZtail() /*readonly*/*zodb.ΔTail {
	return δBtail.δZtail
}

// DB returns database handle that δBtail is using to access ZODB.
func (δBtail *ΔBtail) DB() *zodb.DB {
	return δBtail.db
}


func tracefΔBtail(format string, argv ...interface{}) {
	if traceΔBtail {
		fmt.Printf(format, argv...)
	}
}

func debugfΔBtail(format string, argv ...interface{}) {
	if debugΔBtail {
		fmt.Printf(format, argv...)
	}
}
