// Copyright (C) 2019-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package zdata

// ΔFtail provides ZBigFile-level history tail.
//
// It translates ZODB object-level changes to information about which blocks of
// which ZBigFile were modified, and provides service to query that information.
// See ΔFtail class documentation for usage details.
//
//
// ΔFtail organization
//
// ΔFtail leverages:
//
//	- ΔBtail to track changes to ZBigFile.blktab BTree, and
//	- ΔZtail to track changes to ZBlk objects and to ZBigFile object itself.
//
// then every query merges ΔBtail and ΔZtail data on the fly to provide
// ZBigFile-level result.
//
// Merging on the fly, contrary to computing and maintaining vδF data, is done
// to avoid complexity of recomputing vδF when tracking set changes. Most of
// ΔFtail complexity is, thus, located in ΔBtail, which implements BTree diff
// and handles complexity of recomputing vδB when set of tracked blocks
// changes after new track requests.
//
// Changes to ZBigFile object indicate epochs. Epochs could be:
//
//	- file creation or deletion,
//	- change of ZBigFile.blksize,
//	- change of ZBigFile.blktab to point to another BTree.
//
// Epochs represent major changes to file history where file is assumed to
// change so dramatically, that practically it can be considered to be a
// "whole" change. In particular, WCFS, upon seeing a ZBigFile epoch,
// invalidates all data in corresponding OS-level cache for the file.
//
// The only historical data, that ΔFtail maintains by itself, is history of
// epochs. That history does not need to be recomputed when more blocks become
// tracked and is thus easy to maintain. It also can be maintained only in
// ΔFtail because ΔBtail and ΔZtail does not "know" anything about ZBigFile.
//
// XXX concurrency

import (
	"context"
	"fmt"
	"sort"

	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"

	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/set"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xbtree"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xtail"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xzodb"
)

type setI64 = set.I64
type setOid = set.Oid

// ΔFtail represents tail of revisional changes to files.
//
// It semantically consists of
//
//	[]δF			; rev ∈ (tail, head]
//
// where δF represents a change in files space
//
//	δF:
//		.rev↑
//		{} file ->  {}blk | EPOCH
//
// Only files and blocks explicitly requested to be tracked are guaranteed to
// be present. In particular a block that was not explicitly requested to be
// tracked, even if it was changed in δZ, is not guaranteed to be present in δF.
//
// After file epoch (file creation, deletion, or any other change to file
// object) previous track requests for that file become forgotten and have no
// further effect.
//
// ΔFtail provides the following operations:
//
//   .Track(file, blk, path, zblk)	- add file and block reached via BTree path to tracked set.
//
//   .Update(δZ) -> δF				- update files δ tail given raw ZODB changes
//   .ForgetPast(revCut)			- forget changes ≤ revCut
//   .SliceByRev(lo, hi) -> []δF		- query for all files changes with rev ∈ (lo, hi]
//   .SliceByFileRev(file, lo, hi) -> []δfile	- query for changes of a file with rev ∈ (lo, hi]
//   .BlkRevAt(file, #blk, at) -> blkrev	- query for what is last revision that changed
//						  file[#blk] as of @at database state.
//
// where δfile represents a change to one file
//
//	δfile:
//		.rev↑
//		{}blk | EPOCH
//
// XXX concurrent use.
//
// See also zodb.ΔTail and xbtree.ΔBtail
type ΔFtail struct {
	// ΔFtail merges ΔBtail with history of ZBlk
	δBtail      *xbtree.ΔBtail
	byFile      map[zodb.Oid]*_ΔFileTail // file -> vδf tail
	filesByRoot map[zodb.Oid]setOid      // tree-root -> {} ZBigFile<oid>    as of @head

	// set of files, which are newly tracked and for which byFile[foid].vδE was not yet rebuilt
	trackNew setOid // {}foid

	// set of tracked ZBlk objects reverse-mapped to trees and block numbers
	trackSetZBlk  map[zodb.Oid]*zblkTrack // zblk -> {} root -> {}blk    as of @head
}

// _ΔFileTail represents tail of revisional changes to one file.
type _ΔFileTail struct {
	root zodb.Oid      // .blktab as of @head
	vδE  []_ΔFileEpoch // epochs (changes to ZBigFile object itself)  ; nil if not yet rebuilt
}

// _ΔFileEpoch represent a change to ZBigFile object.
type _ΔFileEpoch struct {
	Rev        zodb.Tid
	oldRoot    zodb.Oid // .blktab was pointing to oldRoot before   ; VDEL if ZBigFile deleted
	newRoot    zodb.Oid // .blktab was changed to point to newRoot  ; ----//----
	oldBlkSize int64    // .blksize was oldBlkSize                  ; -1   if ZBigFile deleted
	newBlkSize int64    // .blksize was changed to newBlkSize       ; ----//----

	// snapshot of trackSetZBlk for this file right before this epoch
	oldTrackSetZBlk map[zodb.Oid]setI64 // {} zblk -> {}blk
}

// zblkTrack keeps information in which root/blocks ZBlk is present as of @head.
type zblkTrack struct {
	inroot map[zodb.Oid]setI64  // {} root -> {}blk
}


// ΔF represents a change in files space.
type ΔF struct {
	Rev    zodb.Tid
	ByFile map[zodb.Oid]*ΔFile // foid -> δfile
}

// ΔFile represents a change to one file.
type ΔFile struct {
	Rev    zodb.Tid
	Epoch  bool   // whether file changed completely
	Blocks setI64 // changed blocks
	Size   bool   // whether file size changed
}


// NewΔFtail creates new ΔFtail object.
//
// Initial tracked set is empty.
// Initial coverage of created ΔFtail is (at₀, at₀].
//
// db will be used by ΔFtail to open database connections to load data from
// ZODB when needed.
func NewΔFtail(at0 zodb.Tid, db *zodb.DB) *ΔFtail {
	return &ΔFtail{
		δBtail:        xbtree.NewΔBtail(at0, db),
		byFile:        map[zodb.Oid]*_ΔFileTail{},
		filesByRoot:   map[zodb.Oid]setOid{},
		trackNew:      setOid{},
		trackSetZBlk:  map[zodb.Oid]*zblkTrack{},
	}
}

// (tail, head] coverage
func (δFtail *ΔFtail) Head() zodb.Tid { return δFtail.δBtail.Head() }
func (δFtail *ΔFtail) Tail() zodb.Tid { return δFtail.δBtail.Tail() }


// ---- Track/rebuild/Update/Forget ----

// Track associates file[blk]@head with tree path and zblk object.
//
// Path root becomes associated with the file, and the path and zblk object become tracked.
// One root can be associated with several files (each provided on different Track calls).
//
// zblk can be nil, which represents a hole.
// blk=-1 should be used for tracking after ZBigFile.Size() query (no zblk is accessed at all).
//
// Objects in path and zblk must be with .PJar().At() == .head
func (δFtail *ΔFtail) Track(file *ZBigFile, blk int64, path []btree.LONode, zblk ZBlk) {
	// XXX locking

	head := δFtail.Head()

	fileAt := file.PJar().At()
	if fileAt != head {
		panicf("file.at (@%s)  !=  δFtail.head (@%s)", fileAt, head)
	}
	if zblk != nil {
		zblkAt := zblk.PJar().At()
		if zblkAt != head {
			panicf("zblk.at (@%s)  !=  δFtail.head (@%s)", zblkAt, head)
		}
	}
	// path.at == head is verified by ΔBtail.Track

	foid := file.POid()
	if blk == -1 {
		blk = xbtree.KeyMax
	}
	δFtail.δBtail.Track(blk, path)

	rootObj := path[0].(*btree.LOBTree)
	root    := rootObj.POid()
	files, ok := δFtail.filesByRoot[root]
	if !ok {
		files = setOid{}
		δFtail.filesByRoot[root] = files
	}
	files.Add(foid)

	δftail, ok := δFtail.byFile[foid]
	if !ok {
		δftail = &_ΔFileTail{root: root, vδE: nil /*will need to be rebuilt to past till tail*/}
		δFtail.byFile[foid] = δftail
		δFtail.trackNew.Add(foid)
	}
	if δftail.root != root {
		// .root can change during epochs, but in between them it must be stable
		panicf("BUG: zfile<%s> root mutated from %s -> %s", foid, δftail.root, root)
	}


	// associate zblk with root, if it was not hole
	if zblk != nil {
		zoid := zblk.POid()
		zt, ok := δFtail.trackSetZBlk[zoid]
		if !ok {
			zt = &zblkTrack{}
			δFtail.trackSetZBlk[zoid] = zt
		}

		inblk, ok := zt.inroot[root]
		if !ok {
			inblk = make(setI64, 1)
			if zt.inroot == nil {
				zt.inroot = make(map[zodb.Oid]setI64)
			}
			zt.inroot[root] = inblk
		}
		inblk.Add(blk)
	}
}

// rebuildAll rebuilds vδE for all files from trackNew requests.
func (δFtail *ΔFtail) rebuildAll() (err error) {
	defer xerr.Contextf(&err, "ΔFtail rebuildAll")
	// XXX locking

	δBtail := δFtail.δBtail
	δZtail := δBtail.ΔZtail()
	db     := δBtail.DB()
	for foid := range δFtail.trackNew {
		δFtail.trackNew.Del(foid)
		δftail := δFtail.byFile[foid]
		err := δftail.rebuild1(foid, δZtail, db)
		if err != nil {
			return err
		}
	}

	return nil
}

// rebuild1IfNeeded rebuilds vδE if there is such need.
//
// it returns corresponding δftail for convenience.
// the only case when vδE actually needs to be rebuilt is when the file just started to be tracked.
func (δFtail *ΔFtail) rebuild1IfNeeded(foid zodb.Oid) (_ *_ΔFileTail, err error) {
	// XXX locking

	δftail := δFtail.byFile[foid]
	if δftail.vδE != nil {
		err = nil
	} else {
		δFtail.trackNew.Del(foid)
		δBtail := δFtail.δBtail
		err = δftail.rebuild1(foid, δBtail.ΔZtail(), δBtail.DB())
	}
	return δftail, err
}

// rebuild1 rebuilds vδE.
func (δftail *_ΔFileTail) rebuild1(foid zodb.Oid, δZtail *zodb.ΔTail, db *zodb.DB) (err error) {
	defer xerr.Contextf(&err, "file<%s>: rebuild", foid)

	// XXX locking
	if δftail.vδE != nil {
		panic("rebuild1: vδE != nil")
	}

	vδE := []_ΔFileEpoch{}
	vδZ := δZtail.Data()
	atPrev := δZtail.Tail()
	for i := 0; i < len(vδZ); i++ {
		δZ := vδZ[i]

		fchanged := false
		for _, oid := range δZ.Changev {
			if oid == foid {
				fchanged = true
				break
			}
		}
		if !fchanged {
			continue
		}

		δ, err := zfilediff(db, foid, atPrev, δZ.Rev)
		if err != nil {
			return err
		}

		if δ != nil {
			δE := _ΔFileEpoch{
				Rev:             δZ.Rev,
				oldRoot:         δ.blktabOld,
				newRoot:         δ.blktabNew,
				oldBlkSize:      δ.blksizeOld,
				newBlkSize:      δ.blksizeNew,
				oldTrackSetZBlk: nil, // nothing was tracked
			}
			vδE = append(vδE, δE)
		}

		atPrev = δZ.Rev
	}

	δftail.vδE = vδE
	return nil
}

// Update updates δFtail given raw ZODB changes.
//
// It returns change in files space that corresponds to δZ.
//
// δZ should include all objects changed by ZODB transaction.
func (δFtail *ΔFtail) Update(δZ *zodb.EventCommit) (_ ΔF, err error) {
	defer xerr.Contextf(&err, "ΔFtail update %s -> %s", δFtail.Head(), δZ.Tid)

	// XXX locking

	// rebuild vδE for newly tracked files
	err = δFtail.rebuildAll()
	if err != nil {
		return ΔF{}, err
	}

	headOld := δFtail.Head()
	δB, err := δFtail.δBtail.Update(δZ)
	if err != nil {
		return ΔF{}, err
	}

	δF := ΔF{Rev: δB.Rev, ByFile: make(map[zodb.Oid]*ΔFile)}

	// take ZBigFile changes into account
	δzfile := map[zodb.Oid]*_ΔZBigFile{} // which tracked ZBigFiles are changed
	for _, oid := range δZ.Changev {
		δftail, ok := δFtail.byFile[oid]
		if !ok {
			continue // not ZBigFile or file is not tracked
		}
		δ, err := zfilediff(δFtail.δBtail.DB(), oid, headOld, δZ.Tid)
		if err != nil {
			return ΔF{}, err
		}
		//fmt.Printf("zfile<%s> diff %s..%s   -> δ: %v\n", oid, headOld, δZ.Tid, δ)

		if δ != nil {
			δzfile[oid] = δ
			δE := _ΔFileEpoch{
				Rev:             δZ.Tid,
				oldRoot:         δ.blktabOld,
				newRoot:         δ.blktabNew,
				oldBlkSize:      δ.blksizeOld,
				newBlkSize:      δ.blksizeNew,
				oldTrackSetZBlk: map[zodb.Oid]setI64{},
			}

			for oid, zt := range δFtail.trackSetZBlk {
				inblk, ok := zt.inroot[δftail.root]
				if ok {
					δE.oldTrackSetZBlk[oid] = inblk
					delete(zt.inroot, δftail.root)
				}
			}

			δftail.root = δE.newRoot
			δftail.vδE  = append(δftail.vδE, δE)
		}
	}

	// take btree changes into account
	//fmt.Printf("δB.ByRoot: %v\n", δB.ByRoot)
	for root, δt := range δB.ByRoot {
		//fmt.Printf("root: %v   δt: %v\n", root, δt)
		files := δFtail.filesByRoot[root]
		// NOTE files might be empty e.g. if a zfile was tracked, then
		// deleted, but the tree referenced by zfile.blktab is still
		// not-deleted, remains tracked and is changed.
		for file := range files {
			δfile, ok := δF.ByFile[file]
			if !ok {
				δfile = &ΔFile{Rev: δF.Rev, Blocks: make(setI64)}
				δF.ByFile[file] = δfile
			}
			for blk /*, δzblk*/ := range δt {
				δfile.Blocks.Add(blk)
			}

			// TODO invalidate .size only if key >= maxkey was changed (size increase),
			// or if on the other hand maxkey was deleted (size decrease).
			//
			// XXX currently we invalidate size on any topology change.
			δfile.Size = true
		}

		// update trackSetZBlk according to btree changes
		for blk, δzblk := range δt {
			if δzblk.Old != xbtree.VDEL {
				ztOld, ok := δFtail.trackSetZBlk[δzblk.Old]
				if ok {
					inblk, ok := ztOld.inroot[root]
					if ok {
						inblk.Del(blk)
					}
				}
			}

			if δzblk.New != xbtree.VDEL {
				ztNew, ok := δFtail.trackSetZBlk[δzblk.New]
				if !ok {
					ztNew = &zblkTrack{}
					δFtail.trackSetZBlk[δzblk.New] = ztNew
				}
				inblk, ok := ztNew.inroot[root]
				if !ok {
					inblk = make(setI64, 1)
					if ztNew.inroot == nil {
						ztNew.inroot = make(map[zodb.Oid]setI64)
					}
					ztNew.inroot[root] = inblk
				}
				inblk.Add(blk)
			}
		}
	}

	// take zblk changes into account
	for _, oid := range δZ.Changev {
		zt, ok := δFtail.trackSetZBlk[oid]
		if !ok {
			continue // not tracked
		}

		for root, inblk := range zt.inroot {
			if len(inblk) == 0 {
				continue
			}
			//fmt.Printf("root: %v   inblk: %v\n", root, inblk)
			files := δFtail.filesByRoot[root]
			for file := range files {
				δfile, ok := δF.ByFile[file]
				if !ok {
					δfile = &ΔFile{Rev: δF.Rev, Blocks: make(setI64)}
					δF.ByFile[file] = δfile
				}

				δfile.Blocks.Update(inblk)
			}
		}
	}

	// if ZBigFile object is changed - it starts new epoch for that file
	for foid, δ := range δzfile {
		δfile, ok := δF.ByFile[foid]
		if !ok {
			δfile = &ΔFile{Rev: δF.Rev}
			δF.ByFile[foid] = δfile
		}
		δfile.Epoch = true
		δfile.Blocks = nil
		δfile.Size = false

		//fmt.Printf("δZBigFile: %v\n", δ)

		// update .filesByRoot
		if δ.blktabOld != xbtree.VDEL {
			files, ok := δFtail.filesByRoot[δ.blktabOld]
			if ok {
				files.Del(foid)
				if len(files) == 0 {
					delete(δFtail.filesByRoot, δ.blktabOld)
				}
			}
		}
		if δ.blktabNew != xbtree.VDEL {
			files, ok := δFtail.filesByRoot[δ.blktabNew]
			if !ok {
				files = setOid{}
				δFtail.filesByRoot[δ.blktabNew] = files
			}
			files.Add(foid)
		}
	}

	//fmt.Printf("-> δF: %v\n", δF)
	return δF, nil
}


// ForgetPast discards all δFtail entries with rev ≤ revCut.
func (δFtail *ΔFtail) ForgetPast(revCut zodb.Tid) {
	δFtail.δBtail.ForgetPast(revCut)

	// XXX locking

	// TODO keep index which file changed epoch where (similarly to ΔBtail),
	// and, instead of scanning all files, trim vδE only on files that is really necessary.
	for _, δftail := range δFtail.byFile {
		δftail.forgetPast(revCut)
	}
}

func (δftail *_ΔFileTail) forgetPast(revCut zodb.Tid) {
	// XXX locking
	icut := 0
	for ; icut < len(δftail.vδE); icut++ {
		if δftail.vδE[icut].Rev > revCut {
			break
		}
	}

	// vδE[:icut] should be forgotten
	if icut > 0 { // XXX workaround for ΔFtail.ForgetPast calling forgetPast on all files
		δftail.vδE = append([]_ΔFileEpoch{}, δftail.vδE[icut:]...)
	}
}

// ---- queries ----

// TODO if needed
// func (δFtail *ΔFtail) SliceByRev(lo, hi zodb.Tid) /*readonly*/ []ΔF

// SliceByFileRev returns history of file changes in (lo, hi] range.
//
// it must be called with the following condition:
//
//	tail ≤ lo ≤ hi ≤ head
//
// the caller must not modify returned slice.
//
// Only tracked blocks are guaranteed to be present.
//
// Note: contrary to regular go slicing, low is exclusive while high is inclusive.
func (δFtail *ΔFtail) SliceByFileRev(zfile *ZBigFile, lo, hi zodb.Tid) /*readonly*/[]*ΔFile {
	foid := zfile.POid()
	//fmt.Printf("\nslice f<%s> (@%s,@%s]\n", foid, lo, hi)
	xtail.AssertSlice(δFtail, lo, hi)

	// XXX locking

	// query .δBtail.SliceByRootRev(file.blktab, lo, hi) +
	// merge δZBlk history with that.

	// merging tree (δT) and Zblk (δZblk) histories into file history (δFile):
	//
	//     δT	────────·──────────────·─────────────────·────────────
	//		        │              │
	//		        ↓              │
	//     δZblk₁	────────────────o───────────────────o─────────────────
	//                                     |
	//                                     ↓
	//     δZblk₂	────────────x────────────────x────────────────────────
	//
	//
	//     δFile	────────o───────o──────x─────x────────────────────────


	δftail, err := δFtail.rebuild1IfNeeded(foid)
	if err != nil {
		panic(err) // XXX
	}

	var vδf []*ΔFile
	// vδfTail returns or creates vδf entry for revision tail
	// tail must be <= all vδf revisions
	vδfTail := func(tail zodb.Tid) *ΔFile {
		if l := len(vδf); l > 0 {
			δfTail := vδf[l-1]
			if δfTail.Rev == tail {
				return δfTail
			}
			if !(tail < δfTail.Rev) {
				panic("tail not ↓")
			}
		}

		δfTail := &ΔFile{Rev: tail, Blocks: setI64{}}
		vδf = append(vδf, δfTail)
		return δfTail
	}

	vδZ := δFtail.δBtail.ΔZtail().SliceByRev(lo, hi)
	iz := len(vδZ) - 1

	// find epoch that covers hi
	vδE := δftail.vδE
	le := len(vδE)
	ie := sort.Search(le, func(i int) bool {
		return hi < vδE[i].Rev
	})
	// vδE[ie]   is next epoch
	// vδE[ie-1] is epoch that covers hi

	// loop through all epochs from hi till lo
	for lastEpoch := false; !lastEpoch ; {
		// current epoch
		var epoch zodb.Tid
		ie--
		if ie < 0 {
			epoch = δFtail.Tail()
		} else {
			epoch = vδE[ie].Rev
		}

		if epoch <= lo {
			epoch = lo
			lastEpoch = true
		}

		var root zodb.Oid // root of blktab in current epoch
		var head zodb.Tid // head] of current epoch coverage
		// state of Zinblk as we are scanning ← current epoch
		// initially corresponds to head of the epoch (= @head for latest epoch)
		Zinblk := map[zodb.Oid]setI64{} // zblk -> which #blk refers to it
		var ZinblkAt zodb.Tid           // Zinblk covers [ZinblkAt,<next δT>)
		if ie+1 == le {
			// head
			root = δftail.root
			head = δFtail.Head()
			for zblk, zt := range δFtail.trackSetZBlk {
				inblk, ok := zt.inroot[root]
				if ok {
					Zinblk[zblk] = inblk.Clone()
				}
			}
		} else {
			δE := vδE[ie+1]
			root = δE.oldRoot
			head = δE.Rev - 1 // TODO better set to exact revision coming before δE.Rev
			for zblk, inblk := range δE.oldTrackSetZBlk {
				Zinblk[zblk] = inblk.Clone()
			}
		}
		//fmt.Printf("Zinblk: %v\n", Zinblk)

		// vδT for current epoch
		vδT := δFtail.δBtail.SliceByRootRev(root, epoch, head) // NOTE @head, not hi
		it := len(vδT) - 1
		if it >= 0 {
			ZinblkAt = vδT[it].Rev
		} else {
			ZinblkAt = epoch
		}

		// merge cumulative vδT(epoch,head] update to Zinblk, so that
		// changes to blocks that were not explicitly requested to be
		// tracked, are present in resulting slice uniformly.
		//
		// For example on
		//
		//	at1   T/B0:a,1:b,2:c   δDø		δ{0,1,2}
		//	at2   δT{0:d,1:e}      δD{c}		δ{0,1,2}
		//	at3   δTø              δD{c,d,e}	δ{0,1,2}
		//	at4   δTø              δD{c,e}		δ{  1,2}
		//
		// if tracked={0}, for (at1,at4] query, changes to 1 should be
		// also all present @at2, @at3 and @at4 - because @at2 both 0
		// and 1 are changed in the same tracked bucket. Note that
		// changes to 2 should not be present at all.
		ZinblkAdj := map[zodb.Oid]setI64{}
		for _, δT := range vδT {
			for blk, δzblk := range δT.KV {
				if δzblk.Old != xbtree.VDEL {
					inblk, ok := ZinblkAdj[δzblk.Old]
					if ok {
						inblk.Del(blk)
					}
				}
				if δzblk.New != xbtree.VDEL {
					inblk, ok := ZinblkAdj[δzblk.New]
					if !ok {
						inblk = setI64{}
						ZinblkAdj[δzblk.New] = inblk
					}
					inblk.Add(blk)
				}
			}
		}
		for zblk, inblkAdj := range ZinblkAdj {
			inblk, ok := Zinblk[zblk]
			if !ok {
				Zinblk[zblk] = inblkAdj
			} else {
				inblk.Update(inblkAdj)
			}
		}

		// merge vδZ and vδT of current epoch
		for ((iz >= 0 && vδZ[iz].Rev > epoch) || it >= 0) {
			// δZ that is covered by current Zinblk
			// -> update δf
			if iz >= 0 && vδZ[iz].Rev > epoch {
				δZ := vδZ[iz]
				if ZinblkAt <= δZ.Rev {
					//fmt.Printf("δZ @%s\n", δZ.Rev)
					for _, oid := range δZ.Changev {
						inblk, ok := Zinblk[oid]
						if ok && len(inblk) != 0 {
							δf := vδfTail(δZ.Rev)
							δf.Blocks.Update(inblk)
						}
					}
					iz--
					continue
				}
			}

			// δT -> adjust Zinblk + update δf
			if it >= 0 {
				δT := vδT[it]
				//fmt.Printf("δT @%s  %v\n", δT.Rev, δT.KV)
				for blk, δzblk := range δT.KV {
					// apply in reverse as we go ←
					if δzblk.New != xbtree.VDEL {
						inblk, ok := Zinblk[δzblk.New]
						if ok {
							inblk.Del(blk)
						}
					}
					if δzblk.Old != xbtree.VDEL {
						inblk, ok := Zinblk[δzblk.Old]
						if !ok {
							inblk = setI64{}
							Zinblk[δzblk.Old] = inblk
						}
						inblk.Add(blk)
					}

					if δT.Rev <= hi {
						δf := vδfTail(δT.Rev)
						δf.Blocks.Add(blk)
						δf.Size = true // see Update
					}
				}

				it--
				if it >= 0 {
					ZinblkAt = vδT[it].Rev
				} else {
					ZinblkAt = epoch
				}
			}
		}

		// emit epoch δf
		if ie >= 0 {
			epoch := vδE[ie].Rev
			if epoch > lo { // it could be <=
				δf := vδfTail(epoch)
				δf.Epoch = true
				δf.Blocks = nil // should be already nil
				δf.Size = false // should be already false
			}
		}
	}

	// vδf was built in reverse order
	// invert the order before returning
	for i,j := 0, len(vδf)-1; i<j; i,j = i+1,j-1 {
		vδf[i], vδf[j] = vδf[j], vδf[i]
	}

	return vδf
}

// BlkRevAt returns last revision that changed file[blk] as of @at database state.
//
// if exact=False - what is returned is only an upper bound for last block revision.
//
// zfile must be any checkout from (tail, head]
// at must ∈ (tail, head]
// blk must be tracked
func (δFtail *ΔFtail) BlkRevAt(ctx context.Context, zfile *ZBigFile, blk int64, at zodb.Tid) (_ zodb.Tid, exact bool) {
	rev, exact, err := δFtail._BlkRevAt(ctx, zfile, blk, at)
	if err != nil {
		panic(err) // XXX
	}
	return rev, exact
}
func (δFtail *ΔFtail) _BlkRevAt(ctx context.Context, zfile *ZBigFile, blk int64, at zodb.Tid) (_ zodb.Tid, exact bool, err error) {
	foid := zfile.POid()
	defer xerr.Contextf(&err, "blkrev f<%s> #%d @%s", foid, blk, at)

	//fmt.Printf("\nblkrev #%d @%s\n", blk, at)

	// assert at ∈ (tail, head]
	tail := δFtail.Tail()
	head := δFtail.Head()
	if !(tail < at && at <= head) {
		panicf("at out of bounds: at: @%s,  (tail, head] = (@%s, @%s]", at, tail, head)
	}

	// assert zfile.at ∈ (tail, head]
	zconn := zfile.PJar()
	zconnAt := zconn.At()
	if !(tail < zconnAt && zconnAt <= head) {
		panicf("zconn.at out of bounds: zconn.at: @%s,  (tail, head] = (@%s, @%s]", zconnAt, tail, head)
	}

	// XXX locking

	δftail, err := δFtail.rebuild1IfNeeded(foid)
	if err != nil {
		return zodb.InvalidTid, false, err
	}

	// find epoch that covers at and associated blktab root/object
	vδE := δftail.vδE
	//fmt.Printf("  vδE: %v\n", vδE)
	l := len(vδE)
	i := sort.Search(l, func(i int) bool {
		return at < vδE[i].Rev
	})
	// vδE[i]   is next epoch
	// vδE[i-1] is epoch that covers at

	// root
	var root zodb.Oid
	if i == l {
		root = δftail.root
	} else {
		root = vδE[i].oldRoot
	}

	// epoch
	var epoch zodb.Tid
	i--
	if i < 0 {
		// i<0 - first epoch (no explicit start) - use δFtail.tail as lo
		epoch = δFtail.Tail()
	} else {
		epoch = vδE[i].Rev
	}

	//fmt.Printf("  epoch: @%s  root: %s\n", epoch, root)

	if root == xbtree.VDEL {
		return epoch, true, nil
	}

	zblk, tabRev, zblkExact, tabRevExact, err := δFtail.δBtail.GetAt(root, blk, at)

	//fmt.Printf("  GetAt  #%d @%s  ->  %s(%v), @%s(%v)\n", blk, at, zblk, zblkExact, tabRev, tabRevExact)
	if err != nil {
		return zodb.InvalidTid, false, err
	}

	if tabRev < epoch {
		tabRev = epoch
		tabRevExact = true
	}

	// if δBtail does not have entry that covers root[blk] - get it
	// through any zconn with .at ∈ (tail, head].
	if !zblkExact {
		xblktab, err := zconn.Get(ctx, root)
		if err != nil {
			return zodb.InvalidTid, false, err
		}
		blktab, err := vBlktab(xblktab)
		if err != nil {
			return zodb.InvalidTid, false, err
		}

		xzblkObj, ok, err := blktab.Get(ctx, blk)
		if err != nil {
			return zodb.InvalidTid, false, err
		}
		if !ok {
			zblk = xbtree.VDEL
		} else {
			zblkObj, err := vZBlk(xzblkObj)
			if err != nil {
				return zodb.InvalidTid, false, fmt.Errorf("blktab<%s>[#%d]: %s", root, blk, err)
			}
			zblk = zblkObj.POid()
		}
	}

	// block was removed
	if zblk == xbtree.VDEL {
		return tabRev, tabRevExact, nil
	}

	// blktab[blk] was changed to point to a zblk @tabRev.
	// blk revision is max rev and when zblk changed last in (rev, at] range.
	zblkRev, zblkRevExact := δFtail.δBtail.ΔZtail().LastRevOf(zblk, at)
	//fmt.Printf("  ZRevOf %s @%s -> @%s, %v\n", zblk, at, zblkRev, zblkRevExact)
	if zblkRev > tabRev {
		return zblkRev, zblkRevExact, nil
	} else {
		return tabRev, tabRevExact, nil
	}
}


// ----------------------------------------


// zfilediff returns direct difference for ZBigFile<foid> old..new .
type _ΔZBigFile struct {
	blksizeOld, blksizeNew int64
	blktabOld,  blktabNew  zodb.Oid
}
func zfilediff(db *zodb.DB, foid zodb.Oid, old, new zodb.Tid) (δ *_ΔZBigFile, err error) {
	txn, ctx := transaction.New(context.TODO()) // TODO - merge in cancel via ctx arg
	defer txn.Abort()
	zconnOld, err := db.Open(ctx, &zodb.ConnOptions{At: old})
	if err != nil {
		return nil, err
	}
	zconnNew, err := db.Open(ctx, &zodb.ConnOptions{At: new})
	if err != nil {
		return nil, err
	}

	a, err1 := zgetFileOrNil(ctx, zconnOld, foid)
	b, err2 := zgetFileOrNil(ctx, zconnNew, foid)
	err = xerr.Merge(err1, err2)
	if err != nil {
		return nil, err
	}

	return diffF(ctx, a, b)
}

// diffF returns direct difference in between two ZBigFile objects.
func diffF(ctx context.Context, a, b *ZBigFile) (δ *_ΔZBigFile, err error) {
	defer xerr.Contextf(&err, "diffF %s %s", xzodb.XidOf(a), xzodb.XidOf(b))

	δ = &_ΔZBigFile{}
	if a == nil {
		δ.blksizeOld = -1
		δ.blktabOld  = xbtree.VDEL
	} else {
		err = a.PActivate(ctx);  if err != nil { return nil, err }
		defer a.PDeactivate()
		δ.blksizeOld = a.blksize
		δ.blktabOld  = a.blktab.POid()
	}

	if b == nil {
		δ.blksizeNew = -1
		δ.blktabNew  = xbtree.VDEL
	} else {
		err = b.PActivate(ctx);  if err != nil { return nil, err }
		defer b.PDeactivate()
		δ.blksizeNew = b.blksize
		δ.blktabNew  = b.blktab.POid()
	}

	// return δ=nil if no change
	if δ.blksizeOld == δ.blksizeNew && δ.blktabOld == δ.blktabNew {
		δ = nil
	}
	return δ, nil
}


// zgetFileOrNil returns ZBigFile corresponding to zconn.Get(oid) .
// if the file does not exist, (nil, ok) is returned.
func zgetFileOrNil(ctx context.Context, zconn *zodb.Connection, oid zodb.Oid) (zfile *ZBigFile, err error) {
	defer xerr.Contextf(&err, "getfile %s@%s", oid, zconn.At())
	xfile, err := xzodb.ZGetOrNil(ctx, zconn, oid)
	if xfile == nil || err != nil {
		return nil, err
	}

	zfile, ok := xfile.(*ZBigFile)
	if !ok {
		return nil, fmt.Errorf("unexpected type: %s", zodb.ClassOf(xfile))
	}
	return zfile, nil
}
